@@ -8,6 +8,7 @@
 */
 
 #include "speakfree.h"
+#include "version.h"
 #include "vat.h"
 
 #define Echo_Base   0		      /* Echo server offset from speaker port */
@@ -445,7 +446,6 @@ static void usage()
     V fprintf(stderr, "           -Vtimeout        Show hostnames that connect\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "by John Walker\n");
-    V fprintf(stderr, "   E-mail: kelvin@fourmilab.ch\n");
     V fprintf(stderr, "   WWW:    http://www.fourmilab.ch/\n");
 }
 
@@ -681,7 +681,7 @@ main(argc, argv)
 #ifndef SHOW_SOCKET
 	if (Debug) {
             fprintf(stderr, "%s: %d bytes read from %s socket.\n", prog, rll,
-                    control ? "control" : "data", rll);
+                    control ? "control" : "data");
 	}
 #endif
 
