@@ -6,6 +6,7 @@
 */
 
 #include "speakfree.h"
+#include "version.h"
 
 #ifdef THREADS
 #include <pthread.h>
@@ -39,6 +40,8 @@ static char *htmlPrivateFile = NULL;  /* HTML private directory file name base *
 static int htmlTime = 1 * 60;	      /* HTML file update time */
 static time_t htmlLast = 0;	      /* HTML last update time */
 static int htmlChange = TRUE;	      /* Change since last HTML update ? */
+static int htmlRefresh = 0;	      /* HTML client-pull refresh interval */
+#define HTML_REFRESH    "<meta http-equiv=\"Refresh\" content=\"%d\">\n"
 static char *message = NULL;	      /* Server information message */
 static int messagel;		      /* Length of server information message */
 
@@ -123,7 +126,7 @@ static char *etime(gmt)
     } else {
 	t = localtime(&clock);
     }
-    sprintf(s, "%02d-%02d %02d:%02d", t->tm_mon + 1, t->tm_mday,
+    sprintf(s, "%d-%02d-%02d %02d:%02d", t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
 	       t->tm_hour, t->tm_min);
     return s;
 }
@@ -195,7 +198,7 @@ static void dumpLwl(fo, lw)
     struct sockaddr_in s;
 
     s.sin_addr.s_addr = lw->naddr;
-    fprintf(fo, "SSRC = %X IP = %s Port %d %s", lw->ssrc, inet_ntoa(s.sin_addr),
+    fprintf(fo, "SSRC = %lX IP = %s Port %d %s", lw->ssrc, inet_ntoa(s.sin_addr),
 		lw->port, ctime(&(lw->ltime)));
     if (lw->cname != NULL) {
         fprintf(fo, "    CNAME = %s\n", lw->cname);
@@ -287,7 +290,7 @@ static int gardol(p, len)
 		  in the HTML file; this allows creation of a non-exported
                   active site file for the system manager's use.  */
 
-static int makeHTML(fname, private)
+static void makeHTML(fname, private)
   char *fname;
   int private;
 {
@@ -304,7 +307,11 @@ static int makeHTML(fname, private)
 
 	LockConn();
 	lw = conn;
-        P("<html>\n<head>\n<title>\nSpeak Freely: Active Sites\n");
+        P("<html>\n<head>\n");
+	if (htmlRefresh > 0) {
+	    fprintf(of, HTML_REFRESH, htmlRefresh);
+	}
+        P("<title>\nSpeak Freely: Active Sites\n");
         P("</title>\n</head>\n\n<body>\n<center>\n<h1>");
         P("Speak Freely: Active Sites</h1>\n<h2>");
         fprintf(of, "As of %s UTC</h2>\n</center>\n<p>\n", etime(TRUE));
@@ -336,23 +343,29 @@ static int makeHTML(fname, private)
 			char ipport[40];
 			struct tm *lt;
 			struct sockaddr_in u;
+			char s[384];
 
-			lt = localtime(&lw->ltime);
+			lt = gmtime(&lw->ltime);
 			u.sin_addr.s_addr = lw->naddr;
                         sprintf(ipport, "%s:%d", inet_ntoa(u.sin_addr), lw->port);
-                        fprintf(of, "\n%-24s %-48s %02d:%02d\n", ipport, lw->cname,
+                        sprintf(s, "\n%-24s %-48s %02d:%02d\n", ipport, lw->cname,
 			    lt->tm_hour, lt->tm_min);
+			outHTML(of, s);
 			if (lw->name != NULL) {
-                            fprintf(of, "%25s%s\n", "", lw->name);
+                            sprintf(s, "%25s%s\n", "", lw->name);
+			    outHTML(of, s);
 			}
 			if (lw->loc != NULL) {
-                            fprintf(of, "%25s%s\n", "", lw->loc);
+                            sprintf(s, "%25s%s\n", "", lw->loc);
+			    outHTML(of, s);
 			}
 			if (lw->phone != NULL) {
-                            fprintf(of, "%25sPhone:  %s\n", "", lw->phone);
+                            sprintf(s, "%25sPhone:  %s\n", "", lw->phone);
+			    outHTML(of, s);
 			}
 			if (lw->email != NULL) {
-                            fprintf(of, "%25sE-mail: %s\n", "", lw->email);
+                            sprintf(s, "%25sE-mail: %s\n", "", lw->email);
+			    outHTML(of, s);
 			}
 		    }
 		    lw = lw->next;
@@ -370,9 +383,7 @@ static int makeHTML(fname, private)
 	if (debugging) {
             fprintf(stderr, "%s: updated %s\n", prog, fn);
 	}
-	return TRUE;
     }
-    return FALSE;
 }
 
 /*  UPDHTML  --  Update HTML if necessary.  */
@@ -649,7 +660,9 @@ static void forwardLwlMessage(zp, rll, addr)
   int rll;			/* Total packet length (starting at zp) */
   struct in_addr addr;		/* IP address of originator of this packet */
 {
-    int i, sock, wasfwd = FALSE;
+#ifndef THREADS
+    int i, sock;
+#endif
 
     zp[0] |= 0x40;		/* Set forwarded packet flag bit */
     bcopy(&addr, zp + rll, sizeof(struct in_addr));
@@ -884,6 +897,7 @@ static void usage()
     V fprintf(stderr, "           -Insec     Interval between HTML updates\n");
     V fprintf(stderr, "           -Mfile     Load server message from file\n");
     V fprintf(stderr, "           -Pport     Listen on given port\n");
+    V fprintf(stderr, "           -Rnsec     Client-pull HTML refresh every nsec seconds\n");
     V fprintf(stderr, "           -U         Print this message\n");
     V fprintf(stderr, "           -V[V]      List connections and disconnects [-VV: all packets]\n");
 #ifdef HEXDUMP
@@ -892,7 +906,6 @@ static void usage()
     V fprintf(stderr, "           -Zpath     Write HTML private directory on base path\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "by John Walker\n");
-    V fprintf(stderr, "   E-mail: kelvin@fourmilab.ch\n");
     V fprintf(stderr, "   WWW:    http://www.fourmilab.ch/\n");
 }
 
@@ -903,7 +916,7 @@ static void usage()
 
 #define addSDES(item, text) add_sdes_item(item, text, &ap)
 
-static int add_sdes_item(item, text, app)
+static void add_sdes_item(item, text, app)
   int item;
   char *text;
   char **app;
@@ -978,7 +991,7 @@ s1:             if (ch == '<') {
 
     l = strlen(prune);
     if (l == 0) {
-	return TRUE;
+	return;
     } else if (l > 48) {
 	int n;
 
@@ -1147,7 +1160,7 @@ static void servicePacket(csock, from)
 			lw->port = from.sin_port;
 			lw->ltime = time(NULL);
 			if (prolix) {
-                            printf("%s: %s SDES %08X\n", prog,
+                            printf("%s: %s SDES %08lX\n", prog,
 				inet_ntoa(from.sin_addr), lw->ssrc);
 			}
 			while (cp < lp) {
@@ -1283,7 +1296,7 @@ static void servicePacket(csock, from)
 		    struct lwl *lr = conn, *llr = NULL;
 
 		    if (prolix) {
-                        printf("%s: %s BYE %08X\n", prog,
+                        printf("%s: %s BYE %08lX\n", prog,
 			    inet_ntoa(from.sin_addr), rp->r.bye.src[0]);
 		    }
 		    while (lr != NULL) {
@@ -1756,6 +1769,10 @@ main(argc, argv)
 		    lwlport = atoi(op + 1);
 		    break;
 
+                case 'R':             /* -Rnsec  --  Interval between HTML client-pull updates */
+		    htmlRefresh = atoi(op + 1);
+		    break;
+
                 case 'U':             /* -U  --  Print usage information */
                 case '?':             /* -?  --  Print usage information */
 		    usage();
