@@ -1,3 +1,5 @@
+/*#define USE_CURSES*/
+#define BYTE_SWAP_DEBUG
 /*
 
 			Speak Freely for Unix
@@ -8,6 +10,7 @@
 */
 
 #include "speakfree.h"
+#include "version.h"
 
 /*  Destination host descriptor.  */
 
@@ -22,6 +25,8 @@ struct destination {
     char vatdeskey[9];		      /* Destination VAT DES key, if any */
     char ideakey[17];		      /* Destination IDEA key, if any */
     char pgpkey[17];		      /* Destination PGP key, if any */
+    char blowfish_spec; 	      /* Nonzero if Blowfish key specified */
+    BF_KEY blowfishkey; 	      /* Destination Blowfish key, if any */
     char *otp;			      /* Key file */
 };
 
@@ -30,6 +35,10 @@ static int sock;		      /* Communication socket */
 static struct destination *dests = NULL, *dtail;
 static int compressing = FALSE;       /* Compress sound: simple 2X */
 static int gsmcompress = TRUE;	      /* GSM compress buffers */
+static int gsm_byte_order_kludge = FALSE; /* Transition to correct byte order for GSM ? */
+#ifdef BYTE_SWAP_DEBUG
+static int gsm_byte_order_debug = FALSE; /* Send GSM length field with swapped bytes */
+#endif
 static int lpccompress = FALSE;       /* LPC compress buffers */
 static int lpc10compress = FALSE;     /* LPC-10 compress buffers */
 static int adpcmcompress = FALSE;     /* ADPCM compress buffers */
@@ -75,6 +84,8 @@ static char currtpkey[9] = "";        /* Current RTP DES key if currtpkey[0] !=
 static char curvatkey[9] = "";        /* Current VAT DES key if currtpkey[0] != 0 */
 static char curideakey[17] = "";      /* Current IDEA key if curideakey[0] != 0 */
 static char curpgpkey[17] = "";       /* Current PGP key if curpgpkey[0] != 0 */
+static char curblowfish_spec = FALSE; /* Nonzero if Blowfish key specified */
+static BF_KEY curblowfishkey;	      /* Blowfish key */
 static char *curotp = NULL;	      /* Key file buffer */
 #ifdef TINY_PACKETS
 static int sound_packet;	      /* Ideal samples/packet */
@@ -99,6 +110,71 @@ extern int HPTermHookChar;
 
 #define ucase(x)    (islower(x) ? toupper(x) : (x))
 
+#ifndef USE_CURSES
+
+#ifndef UNIX420
+#define UNIX5
+#endif
+
+#ifdef UNIX420
+#include <sgtty.h>
+#endif /* UNIX420 */
+
+#ifdef UNIX5
+#include <termio.h>
+static struct termio old_term_params;
+#endif /* UNIX5 */
+
+/* Set raw mode on terminal file.  Basically, get the terminal into a
+   mode in which all characters can be read as they are entered.  CBREAK
+   mode is not sufficient.  */
+
+static void tty_rawmode()
+{
+#ifdef UNIX420
+   struct sgttyb arg;
+
+   ioctl(fileno(stdin), TIOCGETP, &arg);	  /* get basic parameters */
+   arg.sg_flags |= RAW; 	      /* set raw mode */
+   arg.sg_flags &= ~ECHO;	      /* clear echo mode */
+   ioctl(fileno(stdin), TIOCSETP, &arg);	  /* set it */
+#endif /* UNIX420 */
+
+#ifdef UNIX5
+   struct termio term_params;
+
+   ioctl(fileno(stdin), TCGETA, &old_term_params);
+   term_params = old_term_params;
+   term_params.c_iflag &= ~(ICRNL|IXON|IXOFF);	/* no cr translation */
+   term_params.c_iflag &= ~(ISTRIP);   /* no stripping of high order bit */
+   term_params.c_oflag &= ~(OPOST);    /* no output processing */	
+   term_params.c_lflag &= ~(ISIG|ICANON|ECHO); /* raw mode */
+   term_params.c_cc[4] = 1;  /* satisfy read after 1 char */
+   ioctl(fileno(stdin), TCSETAF, &term_params);
+#endif /* UNIX5 */
+/*printf("\n(raw)\n");*/
+}
+
+/* Restore tty mode */
+
+static void tty_normode()
+{
+#ifdef UNIX420
+struct sgttyb arg;
+
+   ioctl(fileno(stdin), TIOCGETP, &arg);	  /* get basic parameters */
+   arg.sg_flags &= ~RAW;	      /* clear raw mode */
+   arg.sg_flags |= ECHO;	      /* set echo mode */
+   ioctl(fileno(stdin), TIOCSETP, &arg);	  /* set it */
+#endif /* UNIX420 */
+
+#ifdef UNIX5
+   ioctl(fileno(stdin), TCSETAF, &old_term_params);
+#endif /* UNIX5 */
+/*printf("\n(cooked)\n");*/
+}
+#endif
+
 /*  ULARM  --  Wrapper for setitimer() that looks like alarm()
 	       but accepts a time in microseconds.  */
 
@@ -140,7 +216,63 @@ static void gsmcomp(sb)
 
     /* Hide original uncompressed buffer length in first 2 bytes of buffer. */
 
-    *((short *) sb->buffer.buffer_val) = sb->buffer.buffer_len;
+    /* Allow me to explain the following if statement.	Speak
+       Freely for Unix prior to release 6.1e contained a bug
+       which caused it to both send (here) and receive
+       (in speaker.c) GSM packets with the embedded length
+       in host byte order.  This went undetected for a long
+       time because all the Unix machines I have access to
+       for testing are big-endian and thus interoperated
+       correctly with one another and the Windows version
+       (which does handle the byte order correctly).  Many
+       Linux and FreeBSD users only speak to one another,
+       and thus never noticed the problem.  It was only when
+       a little-endian Unix machine attempted to use GSM
+       compression in a connection with a Windows or
+       big-endian Unix machine that what came to be called
+       the "deafening static" problem manifested itself.
+       The fix is obvious, but we have to get there in
+       a cautious manner to avoid breaking interoperability
+       during the transition.  So here's the plan: starting
+       with Unix release 6.1e and the corresponding Windows
+       release, the GSM receiving code in both versions
+       will accept the GSM packet length field in
+       *either* byte order (a byte-swapped value in this
+       field can be detected unambiguously because if
+       byte-swapped it will exceed the actual length of
+       the packet, which is otherwise impossible).  Users
+       who install these releases will be able to receive
+       GSM sound from any version of Speak Freely on any
+       platform, present or future.  Version 6.1e for
+       Unix will, however, continue to send the GSM length
+       field in host byte order.  Why?	Because otherwise
+       Linux and FreeBSD users who communicate with one
+       another would both have to install 6.1e before
+       they could communicate once again.  To make the
+       transition smoother, 6.1e offers a "-tc" option,
+       which causes the GSM length field to be sent
+       in the correct byte order.  This option should be
+       set whenever a Unix user wishes to communicate
+       with a Windows user in GSM mode.  It will also
+       fix connections to big-endian Unix versions 
+       prior to 6.1e.  Once a majority of the installed
+       base is on the versions which don't care about
+       byte order in this field, sending in network byte
+       order will be the default.  */
+
+    if (gsm_byte_order_kludge) {
+	*((short *) sb->buffer.buffer_val) = htons((short) sb->buffer.buffer_len);
+    } else {
+	*((short *) sb->buffer.buffer_val) = sb->buffer.buffer_len;
+    }
+#ifdef BYTE_SWAP_DEBUG
+    if (gsm_byte_order_debug) {
+	unsigned char *bsp = (unsigned char *) sb->buffer.buffer_val;
+
+	bsp[0] = sb->buffer.buffer_len & 0xFF;
+	bsp[1] = (sb->buffer.buffer_len >> 8) & 0xFF;
+    }
+#endif
     sb->buffer.buffer_len = l + sizeof(short);
 }
 
@@ -337,6 +469,8 @@ static int addest(host)
     bcopy(curvatkey, d->vatdeskey, 9);
     bcopy(curideakey, d->ideakey, 17);
     bcopy(curpgpkey, d->pgpkey, 17);
+    d->blowfish_spec = curblowfish_spec;
+    bcopy(&curblowfishkey, &(d->blowfishkey), sizeof(BF_KEY));
     d->otp = curotp;
     d->time_to_live = ttl;
     d->name.sin_family = AF_INET;
@@ -497,7 +631,8 @@ static int sendpkt(sb)
 	    }
 	}
 #endif
-	if (d->deskey[0] || d->ideakey[0] || d->pgpkey[0] || d->otp != NULL) {
+	if (d->deskey[0] || d->ideakey[0] || d->blowfish_spec ||
+	    d->pgpkey[0] || d->otp != NULL) {
 	    soundbuf ebuf;
 	    int i;
 	    LONG slen, plen = 0;
@@ -597,6 +732,29 @@ static int sendpkt(sb)
 		ebuf.compression |= fEncIDEA;
 	    }
 
+	    /* Blowfish encryption. */
+
+	    if (d->blowfish_spec) {
+		unsigned char iv[8];
+
+		bzero(iv, sizeof(iv));
+
+                /* If we're Blowfish encrypting we must round the size of
+		   the data to be sent to be a multiple of 8 so that
+		   the entire Blowfish frame is sent. */
+
+		slen = (slen + 7) & (~7);
+		pktlen = slen + (sizeof(struct soundbuf) - BUFL);
+		if (debugging) {
+                    fprintf(stderr, "Encrypting %d bytes with Blowfish key.\r\n",
+			    slen);
+		}
+		BF_cbc_encrypt((unsigned char *) ebuf.buffer.buffer_val,
+			       (unsigned char *) ebuf.buffer.buffer_val,
+			       slen, &(d->blowfishkey), iv, BF_ENCRYPT);
+		ebuf.compression |= fEncBF;
+	    }
+
 	    /* PGP encryption. */
 
 	    if (d->pgpkey[0]) {
@@ -645,7 +803,7 @@ static int sendpkt(sb)
 		ebuf.compression = htonl(ebuf.compression);
 		ebuf.buffer.buffer_len = htonl(ebuf.buffer.buffer_len);
 	    }
-	    for (i = 0; i < (lpc10comp ? robust : 1); i++) {
+	    for (i = 0; i < (lpc10compress ? robust : 1); i++) {
 		if (sendto(sock, (char *) &ebuf, pktlen,
 		    0, (struct sockaddr *) &(d->name), sizeof d->name) < 0) {
                     perror("sending datagram message");
@@ -669,7 +827,7 @@ static int sendpkt(sb)
 		sb->buffer.buffer_len = htonl(sb->buffer.buffer_len);
 	    }
 
-	    for (i = 0; i < (lpc10comp ? robust : 1); i++) {
+	    for (i = 0; i < (lpc10compress ? robust : 1); i++) {
 		if (sendto(sock, (char *) sb, pktlen,
 		    0, (struct sockaddr *) &(d->name), sizeof d->name) < 0) {
                     perror("sending datagram message");
@@ -804,13 +962,20 @@ static void exiting()
 {
     struct destination *d;
 
+    freeaudio();
+
 #ifdef PUSH_TO_TALK
     if (rawmode) {
         fprintf(stderr, "\r      \r");
-	fcntl(fileno(stdin), F_SETFL, 0);
+#ifdef USE_CURSES
 	nocbreak();
 	echo();
 	endwin();
+	fcntl(fileno(stdin), F_SETFL, 0);
+#else
+	tty_normode();
+	fcntl(fileno(stdin), F_SETFL, 0);
+#endif
     }
 #endif
 
@@ -850,7 +1015,6 @@ static void exiting()
 	}
     }
 
-    freeaudio();
     exit(0);
 }
 
@@ -867,6 +1031,72 @@ static void termchar(ch)
     }
 }
 
+/*  CHATCHAR  --  Test for text chat request character.  */
+
+static int chatchar(ch)
+  int ch;
+{
+    if ((!rtp && !vat) && ch == '.') {
+	char s[133];
+
+	if (rawmode) {
+#ifdef USE_CURSES
+	    nocbreak();
+	    echo();
+	    endwin();
+#else
+	    tty_normode();
+#endif
+	}
+	fcntl(fileno(stdin), F_SETFL, 0);
+	if (push) {
+	    talking = 0;
+	}
+
+        fprintf(stderr, "\rChat:  ");
+	while (TRUE) {
+	    if (fgets(s, (sizeof s) - 1, stdin) != NULL) {
+		struct destination *d;
+
+		while (strlen(s) > 0 && isspace(s[strlen(s) - 1])) {
+		    s[strlen(s) - 1] = 0;
+		}
+		for (d = dests; d != NULL; d = d->dnext) {
+		    char v[256];
+		    int l;
+
+		    l = rtp_make_app(v, ssrc, TRUE, RTCP_APP_TEXT_CHAT, s);
+		    /* Set Speak Freely protocol flag in packet */
+		    v[0] = (v[0] & 0x3F) | (1 << 6);
+		    if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
+				 v, l, TRUE) < 0) {
+                        perror("sending text chat packet");
+		    }
+		}
+/*
+fprintf(stderr, "\nOut of chat mode, input(%d) = \"%s\"\n", strlen(s), s);
+*/
+		fprintf(stderr, talking ? (squelch > 0 ?
+                        "\rQuiet: " : "\rTalk:   ") : "\rPause: ");
+
+		if (rawmode) {
+#ifdef USE_CURSES
+		    initscr();
+		    noecho();
+		    cbreak();
+#else
+		    tty_rawmode();
+#endif
+		}
+		break;
+	    }
+	}
+	fcntl(fileno(stdin), F_SETFL, talking ? O_NDELAY : 0);
+	return TRUE;
+    }
+    return FALSE;
+}
+
 /*  SENDFILE  --  Send a file or, if the file name is NULL or a
 		  single period, send real-time sound input. */
 
@@ -1128,22 +1358,30 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 	    int l;
 
 	    fprintf(stderr,
-                "Space bar switches talk/pause, Esc or \"q\" to quit\nPause: ");
+                "Space bar switches talk/pause, Esc or \"q\" to quit, \".\" for chat text\nPause: ");
 	    fflush(stderr);
+#ifdef USE_CURSES
 	    initscr();
 	    noecho();
 	    cbreak();
+#else
+	    tty_rawmode();
+#endif
 	    rawmode = TRUE;
 	    while (TRUE) {
-		if (l = read(fileno(stdin), &c, 1) == 1 || errno != EINTR) {
+		while (TRUE) {
+		    if (l = read(fileno(stdin), &c, 1) == 1 || errno != EINTR) {
+			break;
+		    }
+		}   
+		if (l == 1 && !chatchar(c)) {
 		    break;
 		}
-	    }	
+	    }
 	    if (l != 1) {
                 perror("waiting for first Pause/Talk character");
 	    }
 	    termchar(c);
-/*          fprintf(stderr, "\rTalk:  "); */
             fprintf(stderr, squelch > 0 ? "\rQuiet:  " : "\rTalk:   ");
 	    fcntl(fileno(stdin), F_SETFL, O_NDELAY);
 	    talking = TRUE;
@@ -1194,6 +1432,9 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 #ifdef HEWLETT_PACKARD
 		    c = HPTermHookChar;
 #endif
+		    if (rlen > 0 && chatchar(c)) {
+			continue;
+		    }
 		    termchar(c);
 		    talking = !talking;
 		    fflush(stderr);
@@ -1224,11 +1465,6 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 		    }
 		    break;
 		} else {
-/*
-if (rlen == -1) {
-perror("Read poll"); fflush(stderr);
-}
-*/
 		    if (rlen == -1 && errno == EINTR) {
 			continue;
 		    }
@@ -1446,6 +1682,7 @@ static void usage()
     V fprintf(stderr, "           -A         Always transmit unless squelched\n");
     V fprintf(stderr, "     *     -B         Push to talk using keyboard\n");
 #endif
+    V fprintf(stderr, "           -BFkey     Blowfish encrypt with key\n");
     V fprintf(stderr, "           -C         Compress subsequent sound\n");
     V fprintf(stderr, "           -D         Enable debug output\n");
     V fprintf(stderr, "           -E[key]    Emit session key string\n");
@@ -1465,12 +1702,15 @@ static void usage()
     V fprintf(stderr, "           -RTP       Use Internet Real-Time Protocol\n");
     V fprintf(stderr, "           -Sn,t      Squelch at level n (0-32767), timeout t milliseconds\n");
     V fprintf(stderr, "     *     -T         Telephone (GSM) compression\n");
+    V fprintf(stderr, "           -TC        GSM with corrected x86 byte order\n");
+#ifdef BYTE_SWAP_DEBUG
+    V fprintf(stderr, "           -TD        Debug: send GSM with swapped byte order\n");
+#endif
     V fprintf(stderr, "           -U         Print this message\n");
     V fprintf(stderr, "           -VAT       Use VAT protocol\n");
     V fprintf(stderr, "           -Z\"user..\" Send PGP session key for user(s)\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "by John Walker\n");
-    V fprintf(stderr, "   E-mail: kelvin@fourmilab.ch\n");
     V fprintf(stderr, "   WWW:    http://www.fourmilab.ch/\n");
 }
 
@@ -1590,8 +1830,30 @@ main(argc, argv)
 		    break;
 
                 case 'B':             /* -B  -- Push to talk (button) */
-		    if (isatty(fileno(stdin))) {
-			push = TRUE;
+				      /* -BFkey  -- Set Blowfish key */ 
+                    if (ucase(op[1]) == 'F') {
+			if (strlen(op + 2) == 0) {
+			    curblowfish_spec = FALSE;
+			} else {
+			    unsigned char bfvec[16];
+
+			    curblowfish_spec = TRUE;
+			    MD5Init(&md5c);
+			    MD5Update(&md5c, op + 2, strlen(op + 2));
+			    MD5Final(bfvec, &md5c);
+			    BF_set_key(&(curblowfishkey), 16, bfvec);
+			    if (debugging) {
+                                fprintf(stderr, "Blowfish key:");
+				for (j = 0; j < 16; j++) {
+                                    fprintf(stderr, " %02X", (bfvec[j] & 0xFF));
+				}
+                                fprintf(stderr, "\n");
+			    }
+			}
+		    } else {
+			if (isatty(fileno(stdin))) {
+			    push = TRUE;
+			}
 		    }
 		    break;
 #endif
@@ -1798,8 +2060,17 @@ main(argc, argv)
 		    break;
 
                 case 'T':             /* -T -- Telephone (GSM) compression */
+				      /* -TC -- GSM transition with swapped length byte order */
 		    gsmcompress = TRUE;
 		    lpccompress = lpc10compress = adpcmcompress = FALSE;
+                    if (ucase(op[1]) == 'C') {
+			gsm_byte_order_kludge = TRUE;
+		    }
+#ifdef BYTE_SWAP_DEBUG
+                    if (ucase(op[1]) == 'D') {
+			gsm_byte_order_debug = TRUE;
+		    }
+#endif
 		    break;
 
                 case 'U':             /* -U  --  Print usage information */
@@ -1915,8 +2186,9 @@ main(argc, argv)
 	    }
 
 	    if (rtp || vat) {
-		if (curideakey[0] || curpgpkey[0] || curotp != NULL) {
-                    fprintf(stderr, "IDEA, PGP, and Key File encryption cannot be used\n    with -RTP or -VAT protocols.\n");
+		if (curideakey[0] || curpgpkey[0] ||
+		    curblowfish_spec || curotp != NULL) {
+                    fprintf(stderr, "Blowfish, IDEA, PGP, and Key File encryption cannot be used\n    with -RTP or -VAT protocols.\n");
 		    return 2;
 		}
 #ifdef ZZZZ
