@@ -8,6 +8,7 @@
 */
 
 #include "speakfree.h"
+#include "version.h"
 
 static int audiok = FALSE;	      /* Audio initialised flag */
 static int audiotime = 0;	      /* Audio timeout counter */
@@ -22,6 +23,8 @@ static char deskey[9] = "";           /* DES key, if any */
 static char rtpdeskey[9] = "";        /* RTP DES key, if any */
 static char vatdeskey[9] = "";        /* VAT DES key, if any */
 static char ideakey[17] = "";         /* IDEA key, if any */
+static char blowfish_spec = FALSE;    /* Nonzero if Blowfish key specified */
+static BF_KEY blowfishkey;	      /* Blowfish key */
 static char *curotp = NULL;	      /* Key file, if any */
 static int whichport = Internet_Port; /* Port to listen on (base address) */
 static int sock;		      /* Data socket */
@@ -32,18 +35,30 @@ static int fromlen;		      /* Length of sending host address */
 static int rll; 		      /* Length of packet read from socket */
 static int showhosts = FALSE;	      /* Show host names that connect */
 static int hosttimeout = 180 * 1000000L; /* Consider host idle after this time */
+static int jitter = 0;		      /* Jitter compensation delay */
+static int jitterwait = 0;	      /* Jitter wait in progress */
+static int jitteridle = 0;	      /* Countdown to reset jitter timer */
+static int jitteridlet = 0;	      /* Jitter idle timeout interval */
+struct jitterbuf {		      /* Jitter queue item */
+    struct jitterbuf *next;
+    LONG length;
+    char value[2];
+};
+static struct jitterbuf *jithead = NULL, *jittail = NULL; /* Jitter queue links */
 static char *prog;		      /* Program name */ 
 static FILE *record = NULL;	      /* Answering machine record file */
 static char *pgppass = NULL;	      /* PGP secret key password */
 static char *replyfile = NULL;	      /* Reply file pathname */
 static char *replycmd = "sfmike -t";  /* Default reply command */
+static char *busysignal = "sleep 10; sfmike %s busy.au"; /* Default busy signal command */
+static int busytimeout = 60 * 1000000L; /* Busy signal timeout interval */
+static int dobusy = FALSE;	      /* Reject calls when busy ? */
 #ifdef HALF_DUPLEX
 static struct in_addr localhost;      /* Our internet address */
 static int halfDuplexMuted = FALSE;   /* Muted by half-duplex transmission */
 #endif
 
 static struct sockaddr_in lookhost;   /* Look who's listening host, if any */
-static int lookwho = FALSE;           /* Contact look who's listening ? */
 static char *sdes = NULL;	      /* RTP SDES packet */
 static int sdesl;		      /* RTP SDES packet length */
 static unsigned long ssrc;	      /* RTP synchronisation source identifier */
@@ -63,20 +78,20 @@ static int lwl_strikes[LWL_MAX_SITES];/* Consecutive failures to send LWL packet
 static int lwl_nsites = 0;	      /* Number of LWL sites published on */
 
 static FILE *facefile = NULL;         /* User's face image */
-static int facePlace = 0;	      /* Placement for next face viewer */
 static int facesDisplayed = 0;	      /* Face viewers currently active */
 
-struct connection *conn = NULL;       /* Chain of current connections */
+static struct connection *conn = NULL;/* Chain of current connections */
 
 #define Debug	((debugforce != 0) ? (debugforce > 0) : debugging)
 
 #define TickTock    (10 * 1000000L)   /* Alarm interval in microseconds */
 #define TockTock    (60 * 1000000L)   /* Alarm interval when no connections open */
+#define JitterTock  (1000000L / 4)    /* Alarm interval for jitter compensation */
 
 #ifdef AUDIO_RELEASE
 #define AudioRelease (AUDIO_RELEASE * 1000000L) /* Release audio when idle this long */
 #else
-#define AudioRelease (2 * TickTock)   /* Release audio when idle this long */
+#define AudioRelease (20 * 1000000L)  /* Release audio when idle this long */
 #endif
 
 static int timerStep = TockTock;      /* Current timer step */
@@ -92,8 +107,18 @@ static void gsmdecomp(sb)
     int i, j, l = 0;
     char *dpx = ((char *) sb->buffer.buffer_val) + sizeof(short);
     static char dcb[BUFL];
-    short declen = *((short *) sb->buffer.buffer_val);
+    short declen = ntohs(*((short *) sb->buffer.buffer_val));
 
+    /* Since some old releases forget to convert this field
+       from network byte order, kludge it back to sanity
+       if this happens.  First we try swapping the bytes.  If
+       that doesn't work, force the standard buffer length.
+       Note that we have to write this out, as ntohs will be
+       a no-op on a big-endian machine. */
+
+    if (declen <= 0 || declen > 1600) {
+	declen = ((declen & 0xFF) << 8) | ((declen >> 8) & 0xFF);
+    }
     if (declen <= 0 || declen > 1600) {
 	declen = 1600;
     }
@@ -115,10 +140,10 @@ static void lpcdecomp(sb, c)
   struct soundbuf *sb;
   struct connection *c;
 {
-    int i, j, l = 0;
+    int i, l = 0;
     char *dpx = ((char *) sb->buffer.buffer_val) + sizeof(short);
     char dcb[BUFL];
-    short declen = ntohl(*((short *) sb->buffer.buffer_val));
+    short declen = ntohs(*((short *) sb->buffer.buffer_val));
 
     if (declen <= 0 || declen > (10 * LPC_FRAME_SIZE)) {
 	declen = 10 * LPC_FRAME_SIZE;
@@ -199,7 +224,7 @@ static char *compressionType(msg)
 static void makeSessionKey(key)
   char *key;
 {
-    int j, k;
+    int j;
     struct MD5Context md5c;
     char md5key[16], md5key1[16];
     char s[1024];
@@ -254,15 +279,35 @@ static void ularm(t)
 static void windtimer()
 {
     if (faceTransferActive > 0) {
+	/* There is an implicit assumption here that FaceFetchInterval
+	   is <= JitterTock.  If this is not the case, jitter compensation
+	   may be too coarse during a face transfer. */
 	timerStep = FaceFetchInterval;
     } else if (audiok || (actives > 0)) {
-	timerStep = TickTock;
+	timerStep = (jitteridle > 0) ? JitterTock : TickTock;
     } else {
 	timerStep = TockTock;
     }
     ularm(timerStep);
 }
 
+/*  ISBUSY  --	Test if a connection is active (item on connection
+		chain with a protocol other than PROTOCOL_UNKNOWN).  */
+
+static int isBusy()
+{
+    struct connection *c;
+
+    c = conn;
+    while (c != NULL) {
+	if (c->con_protocol != PROTOCOL_UNKNOWN) {
+	    return dobusy;
+	}
+	c = c->con_next;
+    }
+    return FALSE;
+}
+
 /*  SENDLWLMESSAGE  --	If enabled, send a message identifying us
                         to each selected Look Who's Listening server.  */
 
@@ -270,7 +315,6 @@ static void sendLwlMessage(dobye)
   int dobye;
 {
     int i, sock;
-    struct sockaddr_in name;
 
     for (i = 0; i < lwl_nsites; i++) {
 	if (lwl_ports[i] >= 0) {
@@ -330,6 +374,67 @@ static void sendLwlMessage(dobye)
     close(sock);
 }
 
+/*  OBTAINAUDIO  --  Attempt to obtain the audio output device.  */
+
+static void obtainaudio()
+{
+    if (!audiok) {
+	if (!soundinit(O_WRONLY)) {
+            perror("opening audio output device");
+            fprintf(stderr, "A common cause of this error is a sound board or\n");
+            fprintf(stderr, "driver which cannot run in full duplex mode (some\n");
+            fprintf(stderr, "boards which are physically capable of full duplex\n");
+            fprintf(stderr, "have drivers which cannot run them in this mode).\n\n");
+            fprintf(stderr, "Try uncommenting the line:\n\n");
+            fprintf(stderr, "    DUPLEX = -DHALF_DUPLEX\n\n");
+            fprintf(stderr, "in the Makefile, then \"make clean\", then \"make\".\n");
+            fprintf(stderr, "This will operate your board in half-duplex mode\n");
+            fprintf(stderr, "and avoid the conflict which probably caused the\n");
+            fprintf(stderr, "error opening audio output.\n");
+            return;                   /* Can't acquire sound device */
+	}
+	audiok = TRUE;
+	if (Debug) {
+            fprintf(stderr, "%s: opening audio device.\n", prog);
+	}
+	windtimer();		      /* Reset timer, if necessary */
+    }
+}
+
+/*  FLUSHJITTER  --  Flush any buffers waiting in the jitter compensation
+		     queue.  */
+
+static void flushjitter()
+{
+    struct jitterbuf *jb;
+
+    /* If any samples are pending in the jitter queue, flush
+       them to the audio output.  */
+
+    if ((jb = jithead) != NULL) {
+	jithead = jittail = NULL;
+
+	if (Debug) {
+            fprintf(stderr, "Flushing jitter queue.\n");
+	}
+	obtainaudio();
+	while (jb != NULL) {
+	    struct jitterbuf *jo = jb;
+
+	    if (Debug) {
+		fprintf(stderr,
+                        "Playing %d samples from jitter queue.\n", jb->length);
+	    }
+	    if (audiok) {
+		audiotime = 0;
+		soundplay(jb->length, jb->value);
+	    }
+	    jb = jb->next;
+	    free(jo);
+	}
+    }
+}
+
 /*  RELEASE  --  Signal-catching function which releases the audio
                  device if we haven't received anything to play in
 		 the last minute. */
@@ -348,6 +453,25 @@ static void release()
 	    c->con_timeout += timerStep;
 	    actives++;
 	}
+
+	/* If the busy signal timer is running, decrement it.
+	   Once it expires the host becomes eligible to attempt
+	   to connect once again.  In essence, the connection
+	   is held open until the busy signal timeout expires, and
+	   is then forced to go idle with the regular timeout
+	   mechanism, making it eligible once again to attempt
+	   connection. */
+
+	if (c->con_busy > 0) {
+	    c->con_busy -= timerStep;
+	    if (c->con_busy <= 0) {
+		c->con_busy = 0;
+		c->con_timeout = hosttimeout;
+		if (Debug) {
+                    fprintf(stderr, "%s: %s %s busy signal timeout expired\n", prog, etime(), c->con_hostname);
+		}
+	    }
+	}
 	n = c->con_next;
 	if (c->con_timeout >= hosttimeout) {
 	    actives--;		      /* Oops--just went inactive */
@@ -380,9 +504,6 @@ static void release()
 		c->face_viewer = 0;
 		if (facesDisplayed > 0) {
 		    facesDisplayed--;
-		    if (facesDisplayed == 0) {
-			facePlace = 0;	  /* Reset face place when no faces are displayed */
-		    }
 		}
 		unlink(c->face_filename);
 		c->face_filename[0] = 0;
@@ -491,13 +612,39 @@ static void release()
 	sendLwlMessage(FALSE);
     }
 
+    /* If sound is waiting in the jitter queue, decrement the
+       jitter timer and flush the queue if it decrements to zero. */
+
+    if (jithead != NULL && jitterwait > 0) {
+	jitterwait -= timerStep;
+	if (jitterwait <= 0) {
+	    jitterwait = 0;
+	    flushjitter();
+	}
+    }
+
+    /* Count down the jitter idle timer.  When it's zero, the
+       first packets of an audio stream will be buffered in
+       the jitter queue. */
+
+    if (jitteridle > 0) {
+	jitteridle -= timerStep;
+	if (jitteridle <= 0) {
+	    jitteridle = 0;
+	    if (Debug) {
+                fprintf(stderr, "%s: jitter idle\n", prog);
+	    }
+	}
+    }
+
     /* If we still own the sound device or have a connection
        open, reset the timer. */
 
     windtimer();
     if (Debug) {
 	time_t t = time(NULL);
-        fprintf(stderr, "Tick: %.2f... %s", timerStep / 1000000.0, ctime(&t));
+        fprintf(stderr, "%s: Tick: %.2f... %s", prog,
+		timerStep / 1000000.0, ctime(&t));
     }
     signal(SIGALRM, release);	      /* Set signal to handle timeout */
 }
@@ -545,6 +692,9 @@ static void viewerterm()
 		  wait3(&status, WNOHANG, 0)
 #endif
 	   ) > 0) {
+if (Debug) {
+  fprintf(stderr, "Child process %d done\n", pid);
+}
 	if (WIFEXITED(status) || WIFSIGNALED(status)) {
 	    struct connection *c = conn;
 
@@ -559,7 +709,6 @@ static void viewerterm()
 		    if (facesDisplayed > 0) {
 			facesDisplayed--;
 			if (facesDisplayed == 0) {
-			    facePlace = 0;    /* Reset face place when no faces are displayed */
 			}
 		    }
 		}
@@ -583,18 +732,7 @@ static void playbuffer(msg, c)
     debugging = (msg->compression & fDebug) ? TRUE : FALSE;
 
     audiotime = 0;		      /* Reset timeout counter */
-    if (!audiok) {
-	if (!soundinit(O_WRONLY)) {
-            perror("opening audio output device");
-            return;                   /* Can't acquire sound device */
-	}
-	audiok = TRUE;
-	if (Debug) {
-            fprintf(stderr, "%s: opening audio device.\n", prog);
-	}
-
-	windtimer();		      /* Reset timer, if necessary */
-    }
+    obtainaudio();		      /* Open audio output if necessary */
 
     if (showhosts && (c->con_compmodes != (fCompressionModes &
 	    msg->compression))) {
@@ -619,7 +757,9 @@ static void playbuffer(msg, c)
     if ((msg->compression & fSetDest) && !noring) {
 	sounddest((msg->compression & fDestJack) ? 1 : 0);
 	if (!(msg->compression & fDestJack)) {
-	    soundplayvol(50);	      /* Make sure volume high enough */
+	    if (audiok) {
+		soundplayvol(50);     /* Make sure volume high enough */
+	    }
 	}
     }
 
@@ -651,6 +791,18 @@ static void playbuffer(msg, c)
 	close_idea();
     }
 
+    if ((msg->compression & fEncBF) && blowfish_spec) {
+	unsigned char iv[8];
+	LONG slen = (len + 7) & (~7);
+
+	bzero(iv, sizeof(iv));
+	if (Debug) {
+            fprintf(stderr, "%s: decrypting %d bytes with Blowfish key.\n", prog, slen);
+	}
+	BF_cbc_encrypt((unsigned char *) val, (unsigned char *) val,
+		       slen, &blowfishkey, iv, BF_DECRYPT);
+    }
+
     if ((msg->compression & fEncIDEA) && ideakey[0]) {
 	unsigned short iv[4];
 	LONG slen = (len + 7) & (~7);
@@ -744,7 +896,51 @@ static void playbuffer(msg, c)
 	val = auxbuf;
     }
 
-    soundplay(len, val);
+    /* If the jitter idle time has elapsed, reset the jitter
+       wait to queue the start of this packet sequence. */
+
+    if (jitter > 0 && jitteridle == 0 && jitterwait == 0) {
+	jitterwait = jitter;
+	if (Debug) {
+            fprintf(stderr, "%s: starting %ld millisecond jitter delay.\n", prog, jitterwait / 1000L);
+	}
+    }
+
+    /* Receipt of any sound buffer resets the jitter idle timer. */
+
+    jitteridle = jitteridlet;	      /* Reset jitter idle countdown */
+
+    /* If the jitter wait has not yet expired, add the samples to
+       the queue waiting to be played at the end of the jitter
+       wait.  */
+
+    if (jitterwait > 0) {
+	struct jitterbuf *jb;
+
+	jb = (struct jitterbuf *) malloc(sizeof(struct jitterbuf) + len);
+	if (jb != NULL) {
+	    bcopy(val, jb->value, len);
+	    jb->length = len;
+	    jb->next = NULL;
+	    if (jittail != NULL) {
+		jittail->next = jb;
+	    }
+	    jittail = jb;
+	    if (jithead == NULL) {
+		jithead = jb;
+		windtimer();
+	    }
+	    if (Debug) {
+                fprintf(stderr, "%s: adding %d samples to jitter queue.\n",
+			prog, len);
+	    }
+	}
+    } else {
+	flushjitter();
+	if (audiok) {
+	    soundplay(len, val);
+	}
+    }
     if (record != NULL) {
 	fwrite(val, len, 1, record);
     }
@@ -770,8 +966,11 @@ static void usage()
     V fprintf(stderr, "Usage: %s [options]\n", prog);
     V fprintf(stderr, "Options:\n");
     V fprintf(stderr, "           -A\"file command\" Create answer command in file\n");
+    V fprintf(stderr, "           -B\"command\"    Execute command for busy signal\n");
+    V fprintf(stderr, "           -BFkey           Blowfish decrypt with key\n");
     V fprintf(stderr, "           -D               Force debug output\n");
     V fprintf(stderr, "           -Ikey            IDEA decrypt with key\n");
+    V fprintf(stderr, "           -Jwait,idle      Jitter delay wait and idle in milliseconds\n");
     V fprintf(stderr, "           -Kkey            DES decrypt with key\n");
 #ifdef MULTICAST
     V fprintf(stderr, "           -Mhost/ip        Join multicast to given name or IP address\n");
@@ -790,7 +989,6 @@ static void usage()
     V fprintf(stderr, "           -Z\"phrase\"       Set PGP secret key pass phrase\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "by John Walker\n");
-    V fprintf(stderr, "   E-mail: kelvin@fourmilab.ch\n");
     V fprintf(stderr, "   WWW:    http://www.fourmilab.ch/\n");
 }
 
@@ -928,6 +1126,44 @@ main(argc, argv)
 			}
 			break;
 
+                    case 'B':             /* -B"command"  --  Execute command for busy signal. */
+					  /* -BFkey  -- Set Blowfish key */ 
+                        if (ucase(op[1]) == 'F') {
+			    if (strlen(op + 2) == 0) {
+				blowfish_spec = FALSE;
+			    } else {
+				struct MD5Context md5c;
+				unsigned char bfvec[16];
+
+				blowfish_spec = TRUE;
+				MD5Init(&md5c);
+				MD5Update(&md5c, op + 2, strlen(op + 2));
+				MD5Final(bfvec, &md5c);
+				BF_set_key(&blowfishkey, 16, bfvec);
+				if (Debug) {
+                                    fprintf(stderr, "Blowfish key:");
+				    for (j = 0; j < 16; j++) {
+                                        fprintf(stderr, " %02X", (bfvec[j] & 0xFF));
+				    }
+                                    fprintf(stderr, "\n");
+				}
+			    }
+			} else {
+			    if (strlen(op + 1) > 0) {
+                                if (!strstr(op + 1, "%s")) {
+				    fprintf(stderr,
+  "%s: invalid -B option command: must contain %%s for IP address insertion.\n",
+				      prog);
+				} else {
+				    busysignal = op + 1;
+				    dobusy = TRUE;
+				}
+			    } else {
+				dobusy = TRUE;
+			    }
+			}
+			break;
+
                     case 'I':             /* -Ikey  --  Set IDEA key */
 			if (strlen(op + 1) == 0) {
 			    ideakey[0] = FALSE;
@@ -938,6 +1174,40 @@ main(argc, argv)
 			    MD5Update(&md5c, op + 1, strlen(op + 1));
 			    MD5Final(ideakey + 1, &md5c);
 			    ideakey[0] = TRUE;
+			    if (Debug) {
+                                fprintf(stderr, "IDEA key:");
+				for (j = 0; j < 16; j++) {
+                                    fprintf(stderr, " %02X", (ideakey[j + 1] & 0xFF));
+				}
+                                fprintf(stderr, "\n");
+			    }
+			}
+			break;
+
+                    case 'J':             /* -Jwait,idle --  Set jitter wait and idle times */
+			if (isdigit(*(op + 1))) {
+                            char *cp = strchr(op + 1, ',');
+
+			    if (cp != NULL) {
+				*cp = 0;
+			    }
+			    jitter = atoi(op + 1) * 1000L;
+			    jitteridlet = (cp == NULL) ? (jitter * 2) :
+				(atoi(cp + 1) * 1000L);
+			    if (jitter < 250000) {
+				fprintf(stderr,
+                     "%s: invalid jitter wait %d--reset to 250 milliseconds.\n", 
+				    prog, jitter / 1000);
+				jitter = 250000;
+			    }
+			    if (jitteridlet < 250000) {
+				fprintf(stderr,
+                     "%s: invalid jitter idle time %d--reset to 250 milliseconds.\n", 
+				    prog, jitteridlet / 1000);
+				jitteridlet = 250000;
+			    }
+			} else {
+			    jitteridlet = 2 * (jitter = 1000000L);
 			}
 			break;
 
@@ -1276,7 +1546,7 @@ main(argc, argv)
 
     while (TRUE) {
 	fd_set fdset;
-	int wsock, control;
+	int wsock, control, busyreject;
 
 	FD_ZERO(&fdset);
 	FD_SET(sock, &fdset);
@@ -1301,7 +1571,7 @@ main(argc, argv)
 #ifdef HEXDUMP
 	if (hexdump) {
             fprintf(stderr, "%s: %d bytes read from %s socket.\n", prog, rll,
-                    control ? "control" : "data", rll);
+                    control ? "control" : "data");
 	    xd(stderr, &sb, rll, TRUE);
 	}
 #endif
@@ -1309,6 +1579,7 @@ main(argc, argv)
 	/* See if this connection is active.  If not, initialise a new
 	   connection. */
 
+	busyreject = FALSE;
 	newconn = FALSE;
 	c = conn;
 	while (c != NULL) {
@@ -1358,18 +1629,76 @@ main(argc, argv)
 	    c->con_protocol = PROTOCOL_UNKNOWN;
 	    c->con_rseq = -1;
 	    c->con_reply_current = FALSE;
+	    c->con_busy = 0;
             bcopy("\221\007\311\201", c->con_session, 4);
 	    lpc_init(&c->lpcc);
+	    busyreject = isBusy();
 	}
 
 	if (c != NULL) {
 	    /* Reset connection timeout. */
 	    c->con_timeout = 0;
+
+            /* If we're rejecting this with a busy signal, start
+	       the busy timeout running and dispatch the busy
+	       signal to the host. */
+
+	    if (busyreject && c->con_busy == 0) {
+		pid_t cpid;
+
+		c->con_busy = busytimeout;
+
+		if (audiok) {
+		    soundterm();
+		    audiok = FALSE;
+		    if (Debug) {
+                         fprintf(stderr, "%s: releasing audio before busy signel fork().\n", prog);
+		    }
+		}
+		cpid = fork();
+		if (cpid == 0) {
+		    char s[256];
+
+		    signal(SIGHUP, SIG_DFL);
+		    signal(SIGINT, SIG_DFL);
+		    signal(SIGTERM, SIG_DFL);
+		    signal(SIGALRM, SIG_DFL);
+		    signal(SIGCHLD, SIG_DFL);
+
+		    /* Now we need to close any shared resources
+		       that might have been inherited from the parent
+		       process to avoid their being locked up for the
+		       duration of the child process execution. */
+
+		    close(sock);
+		    if (record != NULL) {
+			fclose(record);
+		    }
+		    if (facefile != NULL) {
+			fclose(facefile);
+		    }
+		    sprintf(s, busysignal, inet_ntoa(from.sin_addr));
+		    if (Debug) {
+                        fprintf(stderr, "%s: sending busy signal with %s\n",
+			    prog, s);
+		    }
+		    system(s);
+		    exit(0);
+		} else if (cpid == (pid_t) -1) {
+                    perror("creating busy signal process");
+		}
+	    }
+
 	    if (newconn) {
 		if (showhosts) {
-                    fprintf(stderr, "%s: %s %s connect\n", prog, etime(), c->con_hostname);
+                    fprintf(stderr, "%s: %s %s %s\n", prog, etime(),
+			    c->con_hostname,
+                            busyreject ? "sending busy signal" : "connect");
 		}
 	    }
+	    if (busyreject) {
+		continue;
+	    }
 
 	    /* Request face data from the other end, starting with
 	       block zero.  If the connection was created itself
@@ -1399,6 +1728,17 @@ main(argc, argv)
 	    continue;
 	}
 
+        /* If we've sent the connection a busy signal, ignore all
+	   packets from it until the busy timeout expires. */
+
+	if (c->con_busy > 0) {
+if (Debug) {
+    fprintf(stderr, "%s: busy--discarding packet from %s.\n", prog, 
+	c->con_hostname);
+}
+	    continue;
+	}
+
 	wasrtp = FALSE;
 
         /* If a DES key is present and we're talking RTP or VAT
@@ -1428,7 +1768,6 @@ main(argc, argv)
 	       to guess whether something we receive is encrypted. */
 
 	    if (!isValidRTCPpacket((unsigned char *) &sb, rll)) {
-		des_cblock key;
 		des_key_schedule sched;
 		des_cblock ivec;
 		int drll = rll;
@@ -1492,9 +1831,10 @@ main(argc, argv)
 	if (control) {
 	    short protocol = PROTOCOL_VATRTP_CRYPT;
 	    unsigned char *p = (unsigned char *) &sb;
+	    unsigned char *apkt;
 	    int proto = (p[0] >> 6) & 3;
 
-	    if (proto == 0) {
+	    if (proto == 0) {	      /* VAT */
 		/* To avoid spoofing by bad encryption keys, require
                    a proper ID message be seen before we'll flip into
 		   VAT protocol. */
@@ -1530,8 +1870,6 @@ main(argc, argv)
 			    uname[0] = 0;
 			    *ep = 0;
 			    while (--i >= 0 && bp < ep) {
-				unsigned long id = *(unsigned long *) bp;
-
 				bp += 4;
                                 strcat(uname, "\t");
 				strcat(uname, (char *) bp);
@@ -1557,6 +1895,7 @@ main(argc, argv)
 
 		    if (p[1] == 2) {
 			c->con_protocol = protocol = PROTOCOL_UNKNOWN;
+			c->con_timeout = hosttimeout - 1;
 			c->con_uname[0] = c->con_email[0] = 0;
 			if (showhosts) {
                             fprintf(stderr, "%s: %s VAT connection closed.\n",
@@ -1564,20 +1903,50 @@ main(argc, argv)
 			}
 		    }
 		}
-	    } else if (proto == RTP_VERSION || proto == 1) {
+
+	    } else if (proto == RTP_VERSION || proto == 1) { /* RTP */
 		if (isValidRTCPpacket((unsigned char *) &sb, rll)) {
 		    protocol = (proto == 1) ? PROTOCOL_SPEAKFREE : PROTOCOL_RTP;
 		    bcopy(p + 4, c->con_session, 4);  /* Save SSRC */
+
                     /* If it's a BYE packet, reset protocol to unknown. */
+
 		    if (isRTCPByepacket((unsigned char *) &sb, rll)) {
 			c->con_protocol = protocol = PROTOCOL_UNKNOWN;
+			c->con_timeout = hosttimeout - 1;
 			c->con_uname[0] = c->con_email[0] = 0;
 			if (showhosts) {
                             fprintf(stderr, "%s: %s %s connection closed.\n",
 				    prog, c->con_hostname,
                                     proto == 1 ? "Speak Freely" : "RTP");
 			}
-		    } else if (showhosts) {
+
+                    /* If it's a text chat message, print it. */
+
+		    } else if (isRTCPAPPpacket((unsigned char *) &sb, rll,
+				    RTCP_APP_TEXT_CHAT, &apkt) && apkt != NULL) {
+			char *ident = c->con_hostname;
+
+			/* To identify the sender, get successively more
+			   personal depending on the information we have at
+			   hand, working down from hostname (which may just
+                           be an IP address if we couldn't resolve the host,
+			   through E-mail address, to user name. */
+
+			if (c->con_email[0] != 0) {
+			    ident = c->con_email;
+			}
+			if (c->con_uname[0] != 0) {
+			    ident = ident = c->con_uname;
+			}
+
+                        printf("%s: %s\n", ident, (char *) (apkt + 12));
+
+                    /* Otherwise, it's presumably an SDES, from which we
+		       should update the user identity information for the
+		       connection. */
+
+		    } else {
 			struct rtcp_sdes_request rp;
 
 			rp.nitems = 4;
@@ -1605,7 +1974,7 @@ main(argc, argv)
 				strcmp(email, c->con_email) != 0) {
 				strcpy(c->con_uname, uname);
 				strcpy(c->con_email, email);
-				if (uname[0]) {
+				if (showhosts && uname[0]) {
                                     fprintf(stderr, "%s: %s", prog, uname);
 				    if (email[0]) {
                                       fprintf(stderr, " (%s)", email);
@@ -1722,10 +2091,19 @@ main(argc, argv)
 	       buffer could inflict on us. */
 
 	    xbl = sb.buffer.buffer_len + (sizeof(struct soundbuf) - BUFL);
-	    if ((sb.compression & (fEncDES | fEncIDEA | fEncPGP)) != 0) {
+
+	    /* If this packet is encrypted with an algorithm which requires
+	       padding the packet to an 8-byte boundary, adjust the actual
+	       content length to account for the padding. */
+
+	    if ((sb.compression & (fEncDES | fEncIDEA | fEncBF | fEncPGP)) != 0) {
 		xbl = ((sb.buffer.buffer_len + 7) & (~7)) +
 		      (sizeof(struct soundbuf) - BUFL);
 	    }
+
+            /* If packet is compressed with LPC-10, compensate for "packet
+               stuffing". */
+
 	    if ((sb.compression & fCompLPC10) && (sb.buffer.buffer_len >= 16)) {
 		xbl -= 16;
 	    }
@@ -1999,8 +2377,7 @@ main(argc, argv)
 
 	    sb.compression = htonl(sb.compression);
 	    sb.buffer.buffer_len = htonl(sb.buffer.buffer_len);
-	    if (sendto(sock, (char *) &sb,
-		(int) ((sizeof(struct soundbuf) - BUFL) + rll),
+	    if (sendto(sock, (char *) &sb, rll,
 		0, (struct sockaddr *) &(name), sizeof name) < 0) {
                 perror("sending datagram message");
 	    }
@@ -2037,7 +2414,6 @@ main(argc, argv)
 	    char cmd[256], f[40], kmd[16];
 	    FILE *kfile;
 	    FILE *pipe;
-	    long flen;
 	    struct MD5Context md5c;
 
 	    MD5Init(&md5c);
@@ -2077,7 +2453,7 @@ main(argc, argv)
 #endif
                     pipe = popen(cmd, "r");
 		    if (pipe == NULL) {
-                        fprintf(stderr, "Unable to open pipe to: %s\n", c);
+                        fprintf(stderr, "Unable to open pipe to: %s\n", cmd);
 		    } else {
 			int lr;
 
